// C Source File
// Created 7/16/2010; 10:35:24 AM

// Solar89: A 68k C program to tokenize text files into 8XP program variables
// These variables can then be transferred to a computer for conversion
// To convert, install and use the Python language program make8xp

// Known bugs:
//	1. If there are 16+ tokens in a line, the program freezes after running
//		a. Somehow, the heap is being corrupted
//		b. Testing says this is NOT caused by an excessive number of callocs or reallocs
//		c. What else changes with increased number of tokens?
//		d. The problem DEFINITELY lies in tokenizeString and/or tokenizePrgm (replacing the findToken call with createNullToken didn't fix it!)

// To-Do List (in no specific order):
//	1. Add #include directives to token file syntax, allowing one token file to include another
//	2. Add code (and interface) to detokenize back into raw text
//	3. Implement a better algorithm for processing token files (for speed)
//	4. Stop wasting handles by allocating working strings (which are only 5 bytes in size)
//	5. Change to use files.h file I/O routines rather than stdio.h routines to decrease program size

// For v0.2: fix bug, maybe add more tokens (depending on how long that bugfix takes...)

// Defined Comments (for shells)
#define COMMENT_STRING         "Tokenize z80 BASIC prgms"
#define COMMENT_PROGRAM_NAME   "Solar89"
#define COMMENT_VERSION_NUMBER 0,1,5,0
#define COMMENT_AUTHORS        "TC01"
#define COMMENT_BW_ICON \
	{0b0000111111110000, \
	 0b0000000011100000, \
	 0b0000000110000000, \
	 0b0000001100000000, \
	 0b0000111000000000, \
	 0b0001111111111000, \
	 0b0000000000000000, \
	 0b0111110001111100, \
	 0b1100011011000110, \
	 0b1100011011000110, \
	 0b1111111011111110, \
	 0b1111111011111110, \
	 0b1100011000000110, \
	 0b1100011000000110, \
	 0b0111110000000110, \
	 0b0000000000000000}

//Include solar89 header, which contains all GCC4TI libraries and functions called from external files
#include "solar89.h"

//Makes a dialog to get input strings, returns boolean (integer) of 0 or 1 depending on success
int getDialogInput(char *string)
{
	int result = 0;
	
	HANDLE dlg = H_NULL;

	//Create a dialog box and add the necessary inputs
	if ((dlg = DialogNewSimple(140, 55)) == H_NULL) 
	{
		DlgMessage("DMA Failure", "Unable to create dialog input box", BT_OK, BT_CANCEL);
		return result;
	}
	DialogAddTitle(dlg, "Solar89", BT_OK, BT_CANCEL);
	DialogAddRequest(dlg, 5, 15, "Text File:", 0, 17, 18);
	DialogAddRequest(dlg, 5, 25, "Output File:", 17, 17, 18);
	DialogAddRequest(dlg, 5, 35, "Token File:", 34, 17, 18);
	
	//Prompt for input
	if (DialogDo(dlg, CENTER, CENTER, string, NULL) == KEY_ENTER)
		result = 1;

	HeapFree(dlg);
	
	//Return result to say we've done it
	return result;
}

//Extracts the actual program from the text file, returns boolean
int getPrgmFromText(char *filename, char *input, unsigned short size)
{
	FILE *source = NULL;

	int result = 0;
	
	//Open the input file and get its text
	if ((source = fopen(filename, "rt")) == NULL)
	{
		DlgMessage("Error", "Unable to open file", BT_OK, BT_NONE);
		return result;
	}

	//In theory we should check that fread() works, but for some reason the check doesn't work
	/*if (fread(input, size, 1, source) != 1)
	{
		DlgMessage("Read Failure", "Unable to read input file", BT_OK, BT_NONE);
		return result;
	}*/
	
	result = 1;
	fread(input, size, 1, source);
	fclose(source);
	
	return result;
}

int writePrgmToVar(char *filename, unsigned char *output, unsigned short size)
{
	unsigned i;
	int bool = 0;
	unsigned length = 0;

	FILE *result = NULL;

	ST_helpMsg("Writing file");

	//Open a new file for writing in binary mode
	if ((result = fopen(filename, "wb")) == NULL)
	{
		DlgMessage("Open Failure", "Unable to open output file", BT_OK, BT_NONE);
		return bool;
	}
	
	//Calculate length of output file
	for (i = 0; i < size; ++i)
	{
		if (output[i] == '\0' && output[i+1] == '\0')
			break;
		length += 1;
	}

	//Write all tokens in the output to file
	fwrite(output, sizeof(unsigned char), length, result);

	//Write the file type
	fputc(0, result);
	fputs("8XP", result);
	fputc(0, result);
	fputc(OTH_TAG, result);

	//Close the file, and if we're successful so far, boolean = true
	fclose(result);
	bool = 1;

	return bool;
}

// Main Function
void _main(void)
{
	char *string = NULL;
	char *inputfile = NULL;
	char *outputfile = NULL;
	char *tokenfile = NULL;
	char *input = NULL;
	char *tokentext = NULL;

	unsigned char *output = NULL;
	unsigned short size = 0;

	int bool = 0;
	int length = 0;
	int numtokens;
	int NUM_TOKENS = 0;
	int argtype;
	int i = 0;
	
	SYM_ENTRY *sym;
	SYM_ENTRY *toksym;
	ESI argptr;
	tokenInfo *tokens = NULL;

  //Allocate memory to pointers
	if ((string = (char *)calloc(60, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for raw input string", BT_OK, BT_NONE);
		return;
	}
	if ((inputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		return;
	}
	if ((outputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		return;
	}
	if ((tokenfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		return;
	}
	
	//Check for arguments passed to the program (arguments are being grabbed successfully, but NOT appended to input string properly...)
	if (ArgCount() == 3)
	{
		InitArgPtr(argptr);
		while ((argtype = GetArgType(argptr)) != END_TAG)
		{
			i += 1;
			if (argtype == STR_TAG)
			{
				switch (i)
				{
					case 1: 
						inputfile = strcpy(inputfile, GetStrnArg(argptr));
						continue;
					case 2: 
						outputfile = strcpy(outputfile, GetStrnArg(argptr));
						continue;
					case 3: 
						tokenfile = strcpy(tokenfile, GetStrnArg(argptr));
						continue;
					default: 
						continue;
				}
			}
			else
			{
				free(string);
				free(inputfile);
				free(outputfile);
				free(tokenfile);
				ER_throw(ER_ARGUMENT);
			}
		}
	}
	//If there are NOT three arguments, get dialog input, if this is impossible, free and quit
	else
	{
		bool = getDialogInput(string);
		if (bool == 0)
		{
			free(string);
			free(inputfile);
			free(outputfile);
			free(tokenfile);
			return;
		}

		//Retrieve file names from raw input
		for (length = 0; length < 17; ++length)
		{
			inputfile[length] = string[length];
			outputfile[length] = string[length + 17];
			tokenfile[length] = string[length + 34];
		}
	}
	
	//Find the size of the input and allocate memory accordingly
	ST_helpMsg("Reading input file");
	sym = SymFindPtr(SYMSTR(inputfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(sym->handle))->Size + 2;
	if ((input = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		return;
	}

	//Get the raw text of the text file
	bool = getPrgmFromText(inputfile, input, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		return;	
	}
	
	//Allocate memory to the output array, based on input size
	if ((output = (unsigned char *)calloc(size, sizeof(unsigned char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		return;
	}

	//Get the size of the token file and allocate memory accordingly
	ST_helpMsg("Initializing tokens");
	toksym = SymFindPtr(SYMSTR(tokenfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(toksym->handle))->Size + 2;
	NUM_TOKENS = (int)size / 10;
	if ((tokens = (tokenInfo *)calloc(NUM_TOKENS, sizeof(tokenInfo))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token info array", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		free(output);
		return;
	}
	if ((tokentext = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for token file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		return;
	}
	
	//Get input from file to be passed to initTokens procedure
	bool = getPrgmFromText(tokenfile, tokentext, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		free(tokentext);
		return;
	}
	
	//Initialize token infos and tokenize (functions in seperate file), if -1 return
	numtokens = initTokens(tokens, tokentext, NUM_TOKENS);
	if (numtokens == -1)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		free(tokentext);
		return;
	}

	//Now, with the gathered token info objects, tokenize the text (add a boolean system here, so we exit before writing file if error?)
	ST_helpMsg("Tokenizing");
	tokenizePrgm(input, output, tokens, numtokens);

	//Write tokens to application variable
	bool = writePrgmToVar(outputfile, output, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		free(tokentext);
		return;
	}
	
	//Free memory
	free(string);
	free(inputfile);
	free(outputfile);
	free(tokenfile);
	free(tokens);
	free(input);
	free(output);
	free(tokentext);

	//Print message and wait for keypress
	ST_helpMsg("Process complete- press any key");
	ngetchx();
}
